# Django
from django.db import models
from django.db.models import Case, Count, Q, Value, When

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class CharacterAuditQuerySet(models.QuerySet):
    # pylint: disable=duplicate-code
    def visible_to(self, user):
        # superusers get all visible
        if user.is_superuser:
            logger.debug("Returning all characters for superuser %s.", user)
            return self

        if user.has_perm("ledger.char_audit_admin_manager"):
            logger.debug("Returning all characters for %s.", user)
            return self

        try:
            char = user.profile.main_character
            assert char
            queries = [models.Q(eve_character__character_ownership__user=user)]

            if user.has_perm("ledger.char_audit_manager"):
                queries.append(
                    models.Q(eve_character__corporation_id=char.corporation_id)
                )

            logger.debug(
                "%s queries for user %s visible chracters.", len(queries), user
            )

            query = queries.pop()
            for q in queries:
                query |= q
            return self.filter(query)
        except AssertionError:
            logger.debug("User %s has no main character. Nothing visible.", user)
            return self.none()

    def annotate_total_update_status_user(self, user):
        """Get the total update status for the given user."""
        char = user.profile.main_character
        assert char

        query = models.Q(eve_character__character_ownership__user=user)

        return self.filter(query).annotate_total_update_status()

    def annotate_total_update_status(self):
        """Get the total update status."""
        # pylint: disable=import-outside-toplevel
        # AA Ledger
        from ledger.models.characteraudit import CharacterAudit

        sections = CharacterAudit.UpdateSection.get_sections()
        num_sections_total = len(sections)
        qs = (
            self.annotate(
                num_sections_total=Count(
                    "ledger_update_status",
                    filter=Q(ledger_update_status__section__in=sections),
                )
            )
            .annotate(
                num_sections_ok=Count(
                    "ledger_update_status",
                    filter=Q(
                        ledger_update_status__section__in=sections,
                        ledger_update_status__is_success=True,
                    ),
                )
            )
            .annotate(
                num_sections_failed=Count(
                    "ledger_update_status",
                    filter=Q(
                        ledger_update_status__section__in=sections,
                        ledger_update_status__is_success=False,
                    ),
                )
            )
            .annotate(
                num_sections_token_error=Count(
                    "ledger_update_status",
                    filter=Q(
                        ledger_update_status__section__in=sections,
                        ledger_update_status__has_token_error=True,
                    ),
                )
            )
            # pylint: disable=no-member
            .annotate(
                total_update_status=Case(
                    When(
                        active=False,
                        then=Value(CharacterAudit.UpdateStatus.DISABLED),
                    ),
                    When(
                        num_sections_token_error=1,
                        then=Value(CharacterAudit.UpdateStatus.TOKEN_ERROR),
                    ),
                    When(
                        num_sections_failed__gt=0,
                        then=Value(CharacterAudit.UpdateStatus.ERROR),
                    ),
                    When(
                        num_sections_ok=num_sections_total,
                        then=Value(CharacterAudit.UpdateStatus.OK),
                    ),
                    When(
                        num_sections_total__lt=num_sections_total,
                        then=Value(CharacterAudit.UpdateStatus.INCOMPLETE),
                    ),
                    default=Value(CharacterAudit.UpdateStatus.IN_PROGRESS),
                )
            )
        )

        return qs

    def disable_characters_with_no_owner(self) -> int:
        """Disable characters which have no owner. Return count of disabled characters."""
        orphaned_characters = self.filter(
            eve_character__character_ownership__isnull=True, active=True
        )
        if orphaned_characters.exists():
            orphans = list(
                orphaned_characters.values_list(
                    "eve_character__character_name", flat=True
                ).order_by("eve_character__character_name")
            )
            orphaned_characters.update(active=False)
            logger.info(
                "Disabled %d characters which do not belong to a user: %s",
                len(orphans),
                ", ".join(orphans),
            )
            return len(orphans)
        return 0


class CharacterAuditManagerBase(models.Manager):
    def get_queryset(self):
        return CharacterAuditQuerySet(self.model, using=self._db)

    @staticmethod
    def visible_eve_characters(user):
        qs = EveCharacter.objects.get_queryset()
        if user.is_superuser:
            logger.debug("Returning all characters for superuser %s.", user)
            return qs.all()

        if user.has_perm("ledger.char_audit_admin_manager"):
            logger.debug("Returning all characters for %s.", user)
            return qs.all()

        try:
            char = user.profile.main_character
            assert char
            queries = [models.Q(character_ownership__user=user)]

            if user.has_perm("ledger.char_audit_manager"):
                queries.append(models.Q(corporation_id=char.corporation_id))

            logger.debug(
                "%s queries for user %s visible chracters.", len(queries), user
            )

            query = queries.pop()
            for q in queries:
                query |= q
            return qs.filter(query)
        except AssertionError:
            logger.debug("User %s has no main character. Nothing visible.", user)
            return qs.none()

    def visible_to(self, user):
        return self.get_queryset().visible_to(user)


CharacterAuditManager = CharacterAuditManagerBase.from_queryset(CharacterAuditQuerySet)



# Django
from django.contrib.auth.models import User

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter

# Alliance Auth (External Libs)
from app_utils.testing import (
    create_user_from_evecharacter,
)

# AA Ledger
from ledger.models.corporationaudit import CorporationAudit, CorporationUpdateStatus
from ledger.tests.testdata.generate_characteraudit import (
    add_auth_character_to_user,
    create_user_from_evecharacter_with_access,
)


def create_corporationaudit(eve_character: EveCharacter, **kwargs) -> CorporationAudit:
    """Create a LedgerAudit Corporation from EveCharacter"""
    params = {
        "corporation_name": eve_character.corporation_name,
        "corporation": eve_character.corporation,
    }
    params.update(kwargs)
    corporation = CorporationAudit(**params)
    corporation.save()
    return corporation


def create_corporation_update_status(
    corporation_audit: CorporationAudit, **kwargs
) -> CorporationUpdateStatus:
    """Create a Update Status for a Character Audit"""
    params = {
        "corporation": corporation_audit,
    }
    params.update(kwargs)
    update_status = CorporationUpdateStatus(**params)
    update_status.save()
    return update_status


def create_corporationaudit_from_user(user: User, **kwargs) -> CorporationAudit:
    """Create a Character Audit from a user"""
    eve_character = user.profile.main_character
    if not eve_character:
        raise ValueError("User needs to have a main character.")

    kwargs.update({"eve_character": eve_character})
    return create_corporationaudit(**kwargs)


def create_corporationaudit_from_evecharacter(
    character_id: int, **kwargs
) -> CorporationAudit:
    """Create a Audit Character from a existing EveCharacter"""

    _, character_ownership = create_user_from_evecharacter_with_access(
        character_id, disconnect_signals=True
    )
    return create_corporationaudit(character_ownership.character, **kwargs)


def add_corporationaudit_corporation_to_user(
    user: User, character_id: int, disconnect_signals: bool = True, **kwargs
) -> CorporationAudit:
    character_ownership = add_auth_character_to_user(
        user,
        character_id,
        disconnect_signals=disconnect_signals,
    )
    return create_corporationaudit(character_ownership.character, **kwargs)



# Generated by Django 4.2.23 on 2025-07-14 09:04

# Django
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("ledger", "0012_alter_corporationwalletjournalentry_division"),
    ]

    operations = [
        migrations.DeleteModel(
            name="Events",
        ),
    ]



"""Forms for app."""

# Django
from django import forms


class ConfirmForm(forms.Form):
    """Form Confirms."""

    character_id = forms.CharField(
        widget=forms.HiddenInput(),
    )

    planet_id = forms.CharField(
        widget=forms.HiddenInput(),
    )



# Django
from django.db import models

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class CorporationAuditQuerySet(models.QuerySet):
    def visible_to(self, user):
        """Returns a queryset of all corps visible to the user."""
        if user.is_superuser:
            logger.debug(
                "Returning all corps for superuser %s.",
                user,
            )
            return self

        if user.has_perm("ledger.corp_audit_admin_manager"):
            logger.debug("Returning all corps for Corp Audit Admin %s.", user)
            return self

        try:
            char = user.profile.main_character
            assert char
            queries = []

            if user.has_perm("ledger.advanced_access"):
                # If the user has advanced access, return all corps from their characters
                corp_ids = char.character_ownership.user.character_ownerships.all().values_list(
                    "character__corporation_id", flat=True
                )
                queries.append(models.Q(corporation__corporation_id__in=corp_ids))

            logger.debug("%s queries for User %s.", len(queries), user)

            if len(queries) == 0:
                return self.none()

            query = queries.pop()
            for q in queries:
                query |= q
            return self.filter(query)
        except AssertionError:
            logger.debug("User %s has no main character. Nothing visible.", user)
            return self.none()

    def manage_to(self, user):
        """Return all corps that the user can manage."""
        if user.is_superuser:
            logger.debug(
                "Returning all corps for superuser %s.",
                user,
            )
            return self

        if user.has_perm("ledger.corp_audit_admin_manager"):
            logger.debug("Returning all corps for Corp Audit Admin %s.", user)
            return self

        try:
            char = user.profile.main_character
            assert char
            queries = []

            if user.has_perm("ledger.corp_audit_manager"):
                # If the user has corp management access, return all corps from their characters
                corp_ids = char.character_ownership.user.character_ownerships.all().values_list(
                    "character__corporation_id", flat=True
                )
                queries.append(models.Q(corporation__corporation_id__in=corp_ids))

            logger.debug("%s queries for User %s.", len(queries), user)

            if len(queries) == 0:
                return self.none()

            query = queries.pop()
            for q in queries:
                query |= q
            return self.filter(query)
        except AssertionError:
            logger.debug("User %s has no main character. Nothing visible.", user)
            return self.none()


class CorporationAuditManagerBase(models.Manager):
    def visible_to(self, user):
        return self.get_queryset().visible_to(user)

    def manage_to(self, user):
        return self.get_queryset().manage_to(user)


CorporationAuditManager = CorporationAuditManagerBase.from_queryset(
    CorporationAuditQuerySet
)



# Generated by Django 4.2.20 on 2025-05-05 05:25

# Django
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("ledger", "0008_alter_characterminingledger_date"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="characteraudit",
            name="last_update_mining",
        ),
        migrations.RemoveField(
            model_name="characteraudit",
            name="last_update_planetary",
        ),
        migrations.RemoveField(
            model_name="characteraudit",
            name="last_update_wallet",
        ),
        migrations.RemoveField(
            model_name="characterplanet",
            name="last_update",
        ),
        migrations.RemoveField(
            model_name="characterplanetdetails",
            name="last_update",
        ),
        migrations.AddField(
            model_name="characterplanetdetails",
            name="character",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="ledger_character_planet_details",
                to="ledger.characteraudit",
            ),
        ),
        migrations.AlterField(
            model_name="characteraudit",
            name="active",
            field=models.BooleanField(default=True),
        ),
        migrations.AlterField(
            model_name="characterminingledger",
            name="character",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="ledger_character_mining",
                to="ledger.characteraudit",
            ),
        ),
        migrations.AlterField(
            model_name="characterplanet",
            name="character",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="ledger_character_planet",
                to="ledger.characteraudit",
            ),
        ),
        migrations.AlterField(
            model_name="characterplanetdetails",
            name="planet",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="ledger_planet_details",
                to="ledger.characterplanet",
            ),
        ),
        migrations.AlterField(
            model_name="characterwalletjournalentry",
            name="character",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="ledger_character_journal",
                to="ledger.characteraudit",
            ),
        ),
        migrations.CreateModel(
            name="CharacterUpdateStatus",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "section",
                    models.CharField(
                        choices=[
                            ("wallet_journal", "Wallet Journal"),
                            ("mining_ledger", "Mining Ledger"),
                            ("planets", "Planets"),
                            ("planets_details", "Planets Details"),
                        ],
                        db_index=True,
                        max_length=32,
                    ),
                ),
                (
                    "is_success",
                    models.BooleanField(db_index=True, default=None, null=True),
                ),
                ("error_message", models.TextField()),
                ("has_token_error", models.BooleanField(default=False)),
                (
                    "last_run_at",
                    models.DateTimeField(
                        db_index=True,
                        default=None,
                        help_text="Last run has been started at this time",
                        null=True,
                    ),
                ),
                (
                    "last_run_finished_at",
                    models.DateTimeField(
                        db_index=True,
                        default=None,
                        help_text="Last run has been successful finished at this time",
                        null=True,
                    ),
                ),
                (
                    "last_update_at",
                    models.DateTimeField(
                        db_index=True,
                        default=None,
                        help_text="Last update has been started at this time",
                        null=True,
                    ),
                ),
                (
                    "last_update_finished_at",
                    models.DateTimeField(
                        db_index=True,
                        default=None,
                        help_text="Last update has been successful finished at this time",
                        null=True,
                    ),
                ),
                (
                    "character",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="ledger_update_status",
                        to="ledger.characteraudit",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.RenameField(
            model_name="corporationwalletdivision",
            old_name="division",
            new_name="division_id",
        ),
        migrations.RemoveField(
            model_name="corporationaudit",
            name="last_update_wallet",
        ),
        migrations.AddField(
            model_name="corporationaudit",
            name="active",
            field=models.BooleanField(default=True),
        ),
        migrations.CreateModel(
            name="CorporationUpdateStatus",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "section",
                    models.CharField(
                        choices=[
                            ("wallet_division", "Divisions"),
                            ("wallet_journal", "Wallet Journal"),
                        ],
                        db_index=True,
                        max_length=32,
                    ),
                ),
                (
                    "is_success",
                    models.BooleanField(db_index=True, default=None, null=True),
                ),
                ("error_message", models.TextField()),
                ("has_token_error", models.BooleanField(default=False)),
                (
                    "last_run_at",
                    models.DateTimeField(
                        db_index=True,
                        default=None,
                        help_text="Last run has been started at this time",
                        null=True,
                    ),
                ),
                (
                    "last_run_finished_at",
                    models.DateTimeField(
                        db_index=True,
                        default=None,
                        help_text="Last run has been successful finished at this time",
                        null=True,
                    ),
                ),
                (
                    "last_update_at",
                    models.DateTimeField(
                        db_index=True,
                        default=None,
                        help_text="Last update has been started at this time",
                        null=True,
                    ),
                ),
                (
                    "last_update_finished_at",
                    models.DateTimeField(
                        db_index=True,
                        default=None,
                        help_text="Last update has been successful finished at this time",
                        null=True,
                    ),
                ),
                (
                    "corporation",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="ledger_corporation_update_status",
                        to="ledger.corporationaudit",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
    ]



# Standard Library
from sys import audit
from unittest.mock import patch

# Django
from django.test import override_settings
from django.utils import timezone

# Alliance Auth (External Libs)
from app_utils.testing import NoSocketsTestCase

# AA Ledger
from ledger.models.general import EveEntity
from ledger.tests.testdata.esi_stub import esi_client_stub
from ledger.tests.testdata.generate_characteraudit import (
    create_characteraudit_from_evecharacter,
)
from ledger.tests.testdata.generate_planets import create_character_planet
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_eveentity import load_eveentity
from ledger.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "ledger.managers.character_planetary_manager"


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(MODULE_PATH + ".esi")
@patch(MODULE_PATH + ".etag_results")
class TestPlanetaryManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()
        load_eveentity()
        cls.audit = create_characteraudit_from_evecharacter(1001)

    def test_update_planets(self, mock_etag, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        mock_etag.side_effect = lambda ob, token, force_refresh=False: ob.results()

        self.audit.update_planets(force_refresh=False)

        self.assertSetEqual(
            set(self.audit.ledger_character_planet.values_list("planet_id", flat=True)),
            {4001, 4002},
        )
        obj = self.audit.ledger_character_planet.get(planet_id=4001)
        self.assertEqual(obj.planet_id, 4001)
        self.assertEqual(obj.upgrade_level, 5)
        self.assertEqual(obj.num_pins, 5)

        obj = self.audit.ledger_character_planet.get(planet_id=4002)
        self.assertEqual(obj.planet_id, 4002)
        self.assertEqual(obj.upgrade_level, 5)
        self.assertEqual(obj.num_pins, 5)


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(MODULE_PATH + ".esi")
@patch(MODULE_PATH + ".etag_results")
class TestPlanetaryDetailsManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()
        load_eveentity()
        cls.audit = create_characteraudit_from_evecharacter(1001)
        cls.planet = create_character_planet(
            characteraudit=cls.audit, planet_id=4001, upgrade_level=5, num_pins=5
        )

    def test_update_planets_details(self, mock_etag, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        mock_etag.side_effect = lambda ob, token, force_refresh=False: ob.results()

        self.audit.update_planets_details(force_refresh=False)

        self.assertSetEqual(
            set(
                self.audit.ledger_character_planet_details.values_list(
                    "planet__planet_id", flat=True
                )
            ),
            {4001},
        )



"""
Corporation Audit Model
"""

# Standard Library
from collections.abc import Callable

# Third Party
from bravado.exception import HTTPInternalServerError

# Django
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger
from esi.errors import TokenError
from esi.models import Token

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__
from ledger.errors import HTTPGatewayTimeoutError, NotModifiedError
from ledger.managers.corporation_audit_manager import CorporationAuditManager
from ledger.managers.corporation_journal_manager import (
    CorporationDivisionManager,
    CorporationWalletManager,
)
from ledger.models.characteraudit import WalletJournalEntry
from ledger.models.general import UpdateSectionResult, UpdateStatus, _NeedsUpdate
from ledger.providers import esi

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class CorporationAudit(models.Model):

    class UpdateSection(models.TextChoices):
        WALLET_DIVISION_NAMES = "wallet_division_names", _("Divisions Names")
        WALLET_DIVISION = "wallet_division", _("Divisions")
        WALLET_JOURNAL = "wallet_journal", _("Wallet Journal")

        @classmethod
        def get_sections(cls) -> list[str]:
            """Return list of section values."""
            return [choice.value for choice in cls]

        @property
        def method_name(self) -> str:
            """Return method name for this section."""
            return f"update_{self.value}"

    corporation_name = models.CharField(max_length=100, null=True, default=None)

    active = models.BooleanField(default=True)

    corporation = models.OneToOneField(
        EveCorporationInfo,
        on_delete=models.CASCADE,
        related_name="ledger_corporationaudit",
    )

    objects = CorporationAuditManager()

    def __str__(self) -> str:
        try:
            return f"{self.corporation.corporation_name} ({self.id})"
        except AttributeError:
            return f"{self.corporation} ({self.id})"

    def update_wallet_division_names(self, force_refresh: bool) -> None:
        return self.ledger_corporation_division.update_or_create_esi_names(
            self, force_refresh=force_refresh
        )

    def update_wallet_division(self, force_refresh: bool) -> None:
        return self.ledger_corporation_division.update_or_create_esi(
            self, force_refresh=force_refresh
        )

    def update_wallet_journal(self, force_refresh: bool) -> None:
        return CorporationWalletJournalEntry.objects.update_or_create_esi(
            self, force_refresh=force_refresh
        )

    # pylint: disable=duplicate-code
    def calc_update_needed(self) -> _NeedsUpdate:
        """Calculate if an update is needed."""
        sections_needs_update = {
            section: True for section in self.UpdateSection.get_sections()
        }
        existing_sections: models.QuerySet[CorporationUpdateStatus] = (
            self.ledger_corporation_update_status.all()
        )
        needs_update = {
            obj.section: obj.need_update()
            for obj in existing_sections
            if obj.section in sections_needs_update
        }
        sections_needs_update.update(needs_update)
        return _NeedsUpdate(section_map=sections_needs_update)

    # pylint: disable=duplicate-code
    def reset_update_status(self, section: UpdateSection) -> "CorporationUpdateStatus":
        """Reset the status of a given update section and return it."""
        update_status_obj: CorporationUpdateStatus = (
            self.ledger_corporation_update_status.get_or_create(
                section=section,
            )[0]
        )
        update_status_obj.reset()
        return update_status_obj

    # pylint: disable=duplicate-code
    def reset_has_token_error(self) -> None:
        """Reset the has_token_error flag for this corporation."""
        self.ledger_corporation_update_status.filter(
            has_token_error=True,
        ).update(
            has_token_error=False,
        )

    # pylint: disable=duplicate-code
    def update_section_if_changed(
        self,
        section: UpdateSection,
        fetch_func: Callable,
        force_refresh: bool = False,
    ):
        """Update the status of a specific section if it has changed."""
        section = self.UpdateSection(section)
        try:
            data = fetch_func(corporation=self, force_refresh=force_refresh)
            logger.debug("%s: Update has changed, section: %s", self, section.label)
        except HTTPInternalServerError as exc:
            logger.debug("%s: Update has an HTTP internal server error: %s", self, exc)
            return UpdateSectionResult(is_changed=False, is_updated=False)
        except NotModifiedError:
            logger.debug("%s: Update has not changed, section: %s", self, section.label)
            return UpdateSectionResult(is_changed=False, is_updated=False)
        except HTTPGatewayTimeoutError as exc:
            logger.debug(
                "%s: Update has a gateway timeout error, section: %s: %s",
                self,
                section.label,
                exc,
            )
            return UpdateSectionResult(is_changed=False, is_updated=False)
        return UpdateSectionResult(
            is_changed=True,
            is_updated=True,
            data=data,
        )

    # pylint: disable=duplicate-code
    def update_section_log(
        self,
        section: UpdateSection,
        is_success: bool,
        is_updated: bool = False,
        error_message: str = None,
    ) -> None:
        """Update the status of a specific section."""
        error_message = error_message if error_message else ""
        defaults = {
            "is_success": is_success,
            "error_message": error_message,
            "has_token_error": False,
            "last_run_finished_at": timezone.now(),
        }
        obj: CorporationUpdateStatus = (
            self.ledger_corporation_update_status.update_or_create(
                section=section,
                defaults=defaults,
            )[0]
        )
        if is_updated:
            obj.last_update_at = obj.last_run_at
            obj.last_update_finished_at = timezone.now()
            obj.save()
        status = "successfully" if is_success else "with errors"
        logger.info("%s: %s Update run completed %s", self, section.label, status)

    # pylint: disable=duplicate-code
    def perform_update_status(
        self, section: UpdateSection, method: Callable, *args, **kwargs
    ) -> UpdateSectionResult:
        """Perform update status."""
        try:
            result = method(*args, **kwargs)
        except Exception as exc:
            # TODO ADD DISCORD NOTIFICATION?
            error_message = f"{type(exc).__name__}: {str(exc)}"
            is_token_error = isinstance(exc, (TokenError))
            logger.error(
                "%s: %s: Error during update status: %s",
                self,
                section.label,
                error_message,
                exc_info=not is_token_error,  # do not log token errors
            )
            self.ledger_corporation_update_status.update_or_create(
                section=section,
                defaults={
                    "is_success": False,
                    "error_message": error_message,
                    "has_token_error": is_token_error,
                    "last_update_at": timezone.now(),
                },
            )
            raise exc
        return result

    @classmethod
    def get_esi_scopes(cls) -> list[str]:
        """Return list of required ESI scopes to fetch."""
        return [
            # General
            "esi-search.search_structures.v1",
            "esi-universe.read_structures.v1",
            "esi-characters.read_corporation_roles.v1",
            # Mining
            "esi-industry.read_corporation_mining.v1",
            # wallets
            "esi-wallet.read_corporation_wallets.v1",
            "esi-markets.read_corporation_orders.v1",
            "esi-industry.read_corporation_jobs.v1",
            "esi-corporations.read_divisions.v1",
        ]

    def get_token(self, scopes, req_roles) -> Token:
        """Get the token for this corporation."""
        if "esi-characters.read_corporation_roles.v1" not in scopes:
            scopes.append("esi-characters.read_corporation_roles.v1")

        char_ids = EveCharacter.objects.filter(
            corporation_id=self.corporation.corporation_id
        ).values("character_id")

        tokens = Token.objects.filter(character_id__in=char_ids).require_scopes(scopes)

        for token in tokens:
            try:
                roles = esi.client.Character.get_characters_character_id_roles(
                    character_id=token.character_id, token=token.valid_access_token()
                ).result()

                has_roles = False
                for role in roles.get("roles", []):
                    if role in req_roles:
                        has_roles = True

                if has_roles:
                    return token
            except TokenError as e:
                logger.error(
                    "Token ID: %s (%s)",
                    token.pk,
                    e,
                )
        return False

    class Meta:
        default_permissions = ()
        permissions = (
            ("corp_audit_manager", "Has Access to own Corporations."),
            ("corp_audit_admin_manager", "Has access to all Corporations."),
        )


class CorporationWalletDivision(models.Model):
    objects = CorporationDivisionManager()

    corporation = models.ForeignKey(
        CorporationAudit,
        on_delete=models.CASCADE,
        related_name="ledger_corporation_division",
    )
    name = models.CharField(max_length=100, null=True, default=None)
    balance = models.DecimalField(max_digits=20, decimal_places=2)
    division_id = models.IntegerField()

    class Meta:
        default_permissions = ()


class CorporationWalletJournalEntry(WalletJournalEntry):
    division = models.ForeignKey(
        CorporationWalletDivision,
        on_delete=models.CASCADE,
        related_name="ledger_corporation_journal",
    )

    objects = CorporationWalletManager()

    def __str__(self):
        return f"Corporation Wallet Journal: RefType: {self.ref_type} - {self.first_party.name} -> {self.second_party.name}: {self.amount} ISK"

    @classmethod
    def get_visible(cls, user):
        corps_vis = CorporationAudit.objects.visible_to(user)
        return cls.objects.filter(division__corporation__in=corps_vis)


class CorporationUpdateStatus(UpdateStatus):
    """A Model to track the status of the last update."""

    corporation = models.ForeignKey(
        CorporationAudit,
        on_delete=models.CASCADE,
        related_name="ledger_corporation_update_status",
    )
    section = models.CharField(
        max_length=32, choices=CorporationAudit.UpdateSection.choices, db_index=True
    )

    def __str__(self) -> str:
        return f"{self.corporation} - {self.section} - {self.is_success}"



"""PvE Views"""

# Standard Library
import json
from decimal import Decimal

# Django
from django.core.cache import cache
from django.core.handlers.wsgi import WSGIRequest
from django.db.models import DecimalField, Q, Sum
from django.utils.translation import gettext as _

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__
from ledger.app_settings import LEDGER_CACHE_ENABLED, LEDGER_CACHE_STALE
from ledger.constants import NPC_ENTITIES
from ledger.helpers.core import LedgerCore, LedgerEntity
from ledger.helpers.ref_type import RefTypeManager
from ledger.models.corporationaudit import (
    CorporationAudit,
    CorporationWalletDivision,
    CorporationWalletJournalEntry,
)

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class CorporationData(LedgerCore):
    """Class to hold character data for the ledger."""

    # pylint: disable=too-many-positional-arguments
    def __init__(
        self,
        request: WSGIRequest,
        corporation: CorporationAudit,
        division_id: int = None,
        year: int = None,
        month: int = None,
        day: int = None,
    ):
        LedgerCore.__init__(self, year, month, day)
        self.request = request
        self.corporation = corporation
        self.division_id = division_id
        self.auth_char_ids = self.auth_character_ids

    def setup_ledger(self, entity: LedgerEntity = None):
        """Setup the Ledger Data for the Corporation."""
        corporation_id = self.corporation.corporation.corporation_id

        # Base queryset filtered by date and corporation division
        base_qs = self._base_journal_queryset()

        if entity is None:
            # No entity specified: show all entries for the corporation (except self-transfers)
            self.journal = base_qs.exclude(
                first_party_id=corporation_id, second_party_id=corporation_id
            )
            # Prepare auxiliary data used by the view
            self.existing_years = self._compute_existing_years()
            self.entities = self._compute_entities()
            return

        # If the entity is the corporation itself and "all" is set, show all entries
        if self.request.GET.get("all", False) and entity.entity_id == corporation_id:
            self.journal = base_qs.exclude(
                first_party_id=corporation_id, second_party_id=corporation_id
            )
            self.entities = self._compute_entities()
            return

        # Regular entity filtering: include any rows where the entity is a first or second party
        character_ids = entity.get_alts_ids_or_self()
        qs = base_qs.filter(
            Q(first_party_id__in=character_ids) | Q(second_party_id__in=character_ids)
        )
        qs = qs.exclude(first_party_id=corporation_id, second_party_id=corporation_id)

        # If the entity is the corporation itself, include NPC transactions too
        if entity.entity_id == corporation_id:
            qs = qs.filter(
                Q(first_party_id__in=NPC_ENTITIES) | Q(second_party_id__in=NPC_ENTITIES)
            )

        # If entity represents a corporation or alliance, exclude auth account character IDs
        # that are not part of the current entity to avoid double counting
        if entity.type in ["alliance", "corporation"]:
            exclude_ids = self.auth_char_ids - set(character_ids)
            qs = qs.exclude(
                Q(first_party_id__in=exclude_ids) | Q(second_party_id__in=exclude_ids)
            )

        self.journal = qs
        self.entities = self._compute_entities()

    def _base_journal_queryset(self):
        """Return the base queryset filtered by the current date range and corporation division."""
        if self.division_id is not None:
            return CorporationWalletJournalEntry.objects.filter(
                self.filter_date,
                division__corporation=self.corporation,
                division=self.division_id,
            )
        return CorporationWalletJournalEntry.objects.filter(
            self.filter_date, division__corporation=self.corporation
        )

    def _compute_entities(self):
        """Return a set of all entity IDs (first and second parties) present in the current journal."""
        return set(self.journal.values_list("second_party_id", flat=True)) | set(
            self.journal.values_list("first_party_id", flat=True)
        )

    def _compute_existing_years(self):
        """Return the available years for journal entries for this corporation."""
        return (
            CorporationWalletJournalEntry.objects.filter(
                division__corporation=self.corporation
            )
            .exclude(date__year__isnull=True)
            .values_list("date__year", flat=True)
            .order_by("-date__year")
            .distinct()
        )

    def create_entity_data(
        self,
        entity: LedgerEntity,
        alts: EveCharacter = None,
    ) -> dict:
        """Create the URL for entity details based on the view type."""
        ids = (
            list(alts.values_list("character__character_id", flat=True))
            if alts is not None
            else [entity.entity_id]
        )

        # Create Alts Dictionary
        alts_dict = {}
        if alts is not None:
            for alt in alts:
                alts_dict[alt.character.character_id] = alt.character.character_name

            # Remove the main character from the alts dictionary only one entry
            if len(alts_dict) == 1:
                alts_dict.pop(entity.entity_id, None)

        used_pks = set()
        bounty = Decimal(0)
        ess = Decimal(0)
        miscellaneous = Decimal(0)
        costs = Decimal(0)

        for pk, rows in list(self.entries.items()):
            for row in rows:
                if row["first_party_id"] in ids or row["second_party_id"] in ids:
                    if RefTypeManager.special_cases(
                        row, ids=ids, account_char_ids=self.auth_char_ids
                    ):
                        continue
                    bounty += row.get("bounty") or Decimal(0)
                    ess += row.get("ess") or Decimal(0)
                    miscellaneous += row.get("miscellaneous") or Decimal(0)
                    costs += row.get("costs") or Decimal(0)
                    used_pks.add(pk)

        # Remove Used Pks from Entries
        # This is to prevent the entries from being used in the future
        for pk in used_pks:
            self.entries.pop(pk, None)

        total = sum([bounty, ess, miscellaneous, costs])

        if total == 0:
            return None

        char_data = {
            "entity": entity,
            "alts": alts_dict,
            "ledger": {
                "bounty": bounty,
                "ess": ess,
                "miscellaneous": miscellaneous,
                "costs": costs,
                "total": total,
            },
            "type": entity.type,
        }

        return char_data

    def generate_ledger_data(self) -> dict:
        """
        Generate the ledger data for the corporation.

        This method processes the journal entries, builds the ledger data,
        and prepares the context for rendering the corporation ledger view.
        """
        ledger = False
        finished_entities = False

        self.setup_ledger()

        journal = self.journal.values(
            "first_party_id", "second_party_id", "pk", "ref_type"
        ).annotate(
            bounty=Sum(
                "amount",
                filter=Q(ref_type__in=RefTypeManager.BOUNTY_PRIZES),
                output_field=DecimalField(),
            ),
            ess=Sum(
                "amount",
                filter=Q(ref_type__in=RefTypeManager.ESS_TRANSFER),
                output_field=DecimalField(),
            ),
            costs=Sum(
                "amount",
                filter=Q(ref_type__in=RefTypeManager.all_ref_types(), amount__lt=0),
                output_field=DecimalField(),
            ),
            miscellaneous=Sum(
                "amount",
                filter=Q(ref_type__in=RefTypeManager.all_ref_types(), amount__gt=0),
                output_field=DecimalField(),
            ),
        )

        # Get the journal hash and cache header
        ledger_hash = self._get_ledger_journal_hash(self.journal.values_list("pk"))
        ledger_header = self._get_ledger_header(
            ledger_args=f"{self.corporation.corporation.corporation_id}_{self.division_id}",
            year=self.year,
            month=self.month,
            day=self.day,
        )
        cache_header = cache.get(
            ledger_header,
            False,
        )
        logger.debug(
            f"Ledger Header: {ledger_header}, Cache Header: {cache_header}, Journal Hash: {ledger_hash}"
        )

        # Check if the journal is up to date
        journal_up_to_date = cache_header == ledger_hash
        ledger_key = self._build_ledger_cache_key(ledger_header)

        # Check if we have newest cached version of the ledger
        if journal_up_to_date and LEDGER_CACHE_ENABLED:
            ledger = cache.get(f"{ledger_key}-data", False)
            finished_entities = cache.get(f"{ledger_key}-finished_entities", False)

        if finished_entities is False or ledger is False:
            # Build the entries from the journal
            self.entries = {}
            for row in journal:
                self.entries.setdefault(row["pk"], []).append(row)

            # Process Auth Accounts first
            ledger, finished_entities = self._process_auth_accounts()
            # Process remaining entities
            self._process_remaining_entities(ledger, finished_entities)
            # Process corporation entity last to ensure it's always included
            self._handle_entity(
                ledger=ledger,
                entity_id=self.corporation.corporation.corporation_id,
                corporation_obj=self.corporation.corporation,
            )
        # Finalize the billboard for the ledger.
        self.create_rattingbar(list(finished_entities))
        self.create_chord(ledger)

        context = self._build_context(ledger=ledger)

        # Create Cache
        cache.set(key=f"{ledger_key}-data", value=ledger, timeout=LEDGER_CACHE_STALE)
        cache.set(
            key=f"{ledger_key}-finished_entities",
            value=finished_entities,
            timeout=LEDGER_CACHE_STALE,
        )
        cache.set(
            key=self._get_ledger_header(
                ledger_args=f"{self.corporation.corporation.corporation_id}_{self.division_id}",
                year=self.year,
                month=self.month,
                day=self.day,
            ),
            value=ledger_hash,
            timeout=None,  # Cache forever until the journal changes
        )
        return context

    def _build_view_data(self, entity_id: int):
        details_kwargs = {
            "viewname": "corporation_details",
            "corporation_id": self.corporation.corporation.corporation_id,
            "entity_id": entity_id,
        }
        if self.division_id is not None:
            details_kwargs["division_id"] = self.division_id
        return details_kwargs

    def _build_view_url(self, entity_id: int):
        """Return the full URL for a corporation view for the given entity id.

        This wraps create_url(**self._build_view_data(...)) so callers can
        request the URL in a single, readable call without losing ordering.
        """
        return self.create_url(**self._build_view_data(entity_id=entity_id))

    # pylint: disable=too-many-arguments
    def _handle_entity(
        self,
        ledger: list,
        entity_id: int,
        character_obj=None,
        corporation_obj=None,
        alts=None,
        add_finished: bool = True,
        finished_ids=None,
    ) -> set:
        """Create entity object, add to ledger if it has data and return IDs to mark finished.

        - ledger: list to append to
        - entity_id: numeric id
        - character_obj / corporation_obj: optional objects to attach to LedgerEntity
        - alts: optional alts queryset passed to create_entity_data
        - add_finished: whether to return ids that should be added to finished_entities
        - finished_ids: explicit IDs (set or iterable) to mark finished (used for accounts)
        """
        details_url = self._build_view_url(entity_id)
        entity_obj = LedgerEntity(
            entity_id,
            character_obj=character_obj,
            corporation_obj=corporation_obj,
            details_url=details_url,
        )
        char_data = self.create_entity_data(entity=entity_obj, alts=alts)
        if char_data is None:
            return set()
        ledger.append(char_data)
        if not add_finished:
            return set()
        if finished_ids is not None:
            return set(finished_ids)
        return {entity_id}

    def _process_auth_accounts(self):
        """Process Auth Account information for the ledger."""
        ledger = []
        finished_entities = set()
        for account in self.auth_accounts:
            alts = account.user.character_ownerships.all()
            existing_alts = set(
                alts.values_list("character__character_id", flat=True)
            ).intersection(self.entities)
            alts = alts.filter(character__character_id__in=existing_alts)
            if not existing_alts:
                continue
            finished_entities.update(
                self._handle_entity(
                    ledger,
                    account.main_character.character_id,
                    character_obj=account.main_character,
                    alts=alts,
                    finished_ids=existing_alts,
                )
            )
        return ledger, finished_entities

    def _process_remaining_entities(self, ledger, finished_entities: set):
        """Process remaining entities for the ledger."""
        remaining_entities = self.entities - finished_entities
        if not remaining_entities:
            return
        for entity_id in remaining_entities:
            if entity_id in NPC_ENTITIES:
                continue
            if entity_id == self.corporation.corporation.corporation_id:
                continue
            finished_entities.update(self._handle_entity(ledger, entity_id))

    def _build_context(self, ledger):
        """Build the context for the ledger view."""
        view = self._build_view_data(
            entity_id=self.corporation.corporation.corporation_id
        )

        context = {
            "title": f"Corporation Ledger - {self.corporation.corporation.corporation_name}",
            "corporation_id": self.corporation.corporation.corporation_id,
            "division_id": self.division_id,
            "billboard": json.dumps(self.billboard.dict.asdict()),
            "ledger": ledger,
            "divisions": CorporationWalletDivision.objects.filter(
                corporation=self.corporation
            ).order_by("division_id"),
            "years": list(self.existing_years),
            "totals": self._calculate_totals(ledger),
            "view": self.create_view_data(**view),
        }
        return context

    def create_rattingbar(self, entities_ids: list = None):
        """Create the ratting bar for the view."""
        if not entities_ids:
            return

        rattingbar_timeline = self.billboard.create_timeline(self.journal)
        rattingbar = (
            rattingbar_timeline.annotate_bounty_income()
            .annotate_ess_income()
            .annotate_miscellaneous()
        )
        self.billboard.create_or_update_results(rattingbar)
        self._build_xy_chart(title=_("Ratting Bar"))

    def create_chord(self, ledger_data: list[dict]):
        """Create the chord chart for the view."""
        if not ledger_data:
            return

        for entry in ledger_data:
            entity_name = entry["entity"].entity_name
            ledger = entry["ledger"]
            self.billboard.chord_add_data(
                chord_from=entity_name,
                chord_to=_("Bounty (Wallet)"),
                value=ledger.get("bounty", 0),
            )
            self.billboard.chord_add_data(
                chord_from=entity_name,
                chord_to=_("ESS (Wallet)"),
                value=ledger.get("ess", 0),
            )
            self.billboard.chord_add_data(
                chord_from=entity_name,
                chord_to=_("Costs (Wallet)"),
                value=abs(ledger.get("costs", 0)),
            )
            self.billboard.chord_add_data(
                chord_from=entity_name,
                chord_to=_("Miscellaneous (Wallet)"),
                value=abs(ledger.get("miscellaneous", 0)),
            )
        self.billboard.chord_handle_overflow()



# Standard Library
from collections import defaultdict
from typing import TYPE_CHECKING

# Django
from django.db import models, transaction
from django.db.models import (
    DecimalField,
    ExpressionWrapper,
    F,
    Q,
    Sum,
    Value,
)
from django.db.models.functions import Coalesce, Round
from django.utils import timezone

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag
from eveuniverse.models import EveSolarSystem, EveType

# AA Ledger
from ledger import __title__
from ledger.app_settings import LEDGER_PRICE_PERCENTAGE
from ledger.decorators import log_timing
from ledger.helpers.etag import etag_results
from ledger.providers import esi

if TYPE_CHECKING:
    # AA Ledger
    from ledger.models.characteraudit import (
        CharacterAudit,
    )
    from ledger.models.general import UpdateSectionResult

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def require_valid_price_percentage(func):
    def wrapper(*args, **kwargs):
        if not isinstance(LEDGER_PRICE_PERCENTAGE, (int, float)):
            raise ValueError("LEDGER_PRICE_PERCENTAGE must be a number")
        if LEDGER_PRICE_PERCENTAGE <= 0:
            raise ValueError("LEDGER_PRICE_PERCENTAGE must be positive")
        return func(*args, **kwargs)

    return wrapper


class CharacterMiningLedgerEntryQueryset(models.QuerySet):
    @require_valid_price_percentage
    def annotate_pricing(self) -> models.QuerySet:
        """Annotate price and total columns."""
        return self.annotate(price=F("price_per_unit")).annotate(
            total=ExpressionWrapper(
                (F("price_per_unit") * F("quantity")) * LEDGER_PRICE_PERCENTAGE,
                output_field=models.DecimalField(),
            ),
        )

    def annotate_mining(self, with_period: bool = False) -> models.QuerySet:
        """Annotate mining columns."""
        qs = self.annotate_pricing()
        values_fields = [
            "character__eve_character__character_id",
            "character__eve_character__character_name",
        ]
        if with_period:
            values_fields.append("period")

        return qs.values(*values_fields).annotate(
            mining_income=Round(
                Coalesce(
                    Sum(F("total")),
                    Value(0),
                    output_field=DecimalField(),
                ),
                precision=2,
            )
        )

    def aggregate_mining(self):
        """Aggregate mining amounts."""
        qs = self.annotate_pricing()
        return qs.aggregate(
            total_amount=Round(
                Coalesce(
                    Sum(F("total")),
                    Value(0),
                    output_field=DecimalField(),
                ),
                precision=2,
            )
        )["total_amount"]

    def aggregate_amounts_information_modal(
        self, amounts: defaultdict, chars_list: list, filter_date: timezone.datetime
    ) -> dict:
        """Generate data template for the ledger character information view."""
        qs = self.filter(Q(character__eve_character__character_id__in=chars_list))
        qs = qs.annotate_pricing()
        qs = qs.aggregate(
            total_amount=Round(
                Coalesce(
                    Sum(F("total"), filter=Q(date__year=filter_date.year)),
                    Value(0),
                    output_field=DecimalField(),
                ),
                precision=2,
            ),
            total_amount_day=Round(
                Coalesce(
                    Sum(F("total"), filter=Q(date__day=filter_date.day)),
                    Value(0),
                    output_field=DecimalField(),
                ),
                precision=2,
            ),
        )

        amounts["mining"]["total_amount"] = qs["total_amount"]
        amounts["mining"]["total_amount_day"] = qs["total_amount_day"]

        return amounts

    def annotate_billboard(self, chars_list: list) -> models.QuerySet:
        """Annotate billboard columns."""
        qs = self.filter(Q(character__eve_character__character_id__in=chars_list))
        return qs.annotate(
            total_amount=Round(
                Coalesce(
                    Sum(F("total")),
                    Value(0),
                    output_field=DecimalField(),
                ),
                precision=2,
            )
        )


class CharacterMiningLedgerEntryManagerBase(models.Manager):
    @log_timing(logger)
    def update_or_create_esi(
        self, character: "CharacterAudit", force_refresh: bool = False
    ) -> "UpdateSectionResult":
        """Update or Create a mining ledger entry from ESI data."""
        return character.update_section_if_changed(
            section=character.UpdateSection.MINING_LEDGER,
            fetch_func=self._fetch_esi_data,
            force_refresh=force_refresh,
        )

    def _fetch_esi_data(
        self, character: "CharacterAudit", force_refresh: bool = False
    ) -> None:
        """Fetch mining ledger entries from ESI data."""
        req_scopes = ["esi-industry.read_character_mining.v1"]

        token = character.get_token(scopes=req_scopes)
        mining_obj = esi.client.Industry.get_characters_character_id_mining(
            character_id=character.eve_character.character_id
        )
        mining_items = etag_results(mining_obj, token, force_refresh=force_refresh)
        self._update_or_create_objs(character, mining_items)

    @transaction.atomic()
    def _update_or_create_objs(self, character: "CharacterAudit", objs: list) -> None:
        """Update or Create mining ledger entries from objs data."""
        existings_pks = set(
            self.filter(
                character=character,
                date__gte=timezone.now() - timezone.timedelta(days=30),
            ).values_list("id", flat=True)
        )
        type_ids = set()
        system_ids = set()
        new_events = []
        old_events = []

        for entry in objs:
            type_ids.add(entry.get("type_id"))
            system_ids.add(entry.get("solar_system_id"))
            pk = self.model.create_primary_key(character.pk, entry)
            _e = self.model(
                character=character,
                id=pk,
                date=entry.get("date"),
                type_id=entry.get("type_id"),
                system_id=entry.get("solar_system_id"),
                quantity=entry.get("quantity"),
            )
            if pk in existings_pks:
                old_events.append(_e)
            else:
                new_events.append(_e)

        # Ensure both EveType and EveSolarSystem objects exist before creating mining entries
        EveType.objects.bulk_get_or_create_esi(ids=list(type_ids))
        EveSolarSystem.objects.bulk_get_or_create_esi(ids=list(system_ids))

        if new_events:
            self.bulk_create(new_events, ignore_conflicts=True)

        if old_events:
            self.bulk_update(old_events, fields=["quantity"])

        self._update_mining_price(character)

    def _update_mining_price(self, character: "CharacterAudit") -> None:
        """Update prices for mining ledger entries."""
        # Update EveMarketPrice on a Daily basis
        self.model.update_evemarket_price()

        mining_ledger = character.mining_ledger.filter(price_per_unit__isnull=True)
        logger.debug(
            f"Checking {mining_ledger.count()} mining ledger entries for missing prices."
        )

        updated_entries = []
        for entry in mining_ledger:
            npc_price = entry.get_npc_price()
            if npc_price is not None:
                entry.price_per_unit = npc_price
                updated_entries.append(entry)

        if updated_entries:
            self.bulk_update(updated_entries, fields=["price_per_unit"])

        logger.debug(
            f"Updated prices for {len(updated_entries)}({character.character_name}) mining ledger entries."
        )


CharacterMiningLedgerEntryManager = CharacterMiningLedgerEntryManagerBase.from_queryset(
    CharacterMiningLedgerEntryQueryset
)



"""Generate AllianceAuth test objects from allianceauth.json."""

# Standard Library
import json
from pathlib import Path

# AA Ledger
from ledger.models.general import EveEntity


def _load_eveentity_data():
    with open(Path(__file__).parent / "eveentity.json", encoding="utf-8") as fp:
        return json.load(fp)


_entities_data = _load_eveentity_data()


def load_eveentity():
    EveEntity.objects.all().delete()
    for character_info in _entities_data.get("EveEntity"):
        EveEntity.objects.create(
            eve_id=character_info.get("eve_id"),
            name=character_info.get("name"),
            category=character_info.get("category"),
        )



# Generated by Django 4.2.11 on 2025-05-26 06:53

# Django
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("ledger", "0011_create_update_status_existings"),
    ]

    operations = [
        migrations.AlterField(
            model_name="corporationwalletjournalentry",
            name="division",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="ledger_corporation_journal",
                to="ledger.corporationwalletdivision",
            ),
        ),
    ]



# Django
from django.test import TestCase
from django.utils import timezone

# AA Ledger
from ledger.models.characteraudit import CharacterAudit, CharacterUpdateStatus
from ledger.tests.testdata.generate_characteraudit import (
    add_characteraudit_character_to_user,
    create_update_status,
    create_user_from_evecharacter_with_access,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_eveentity import load_eveentity
from ledger.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "ledger.managers.character_audit_manager"


class TestCharacterAuditManager(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()
        load_eveentity()
        cls.user, cls.character_ownership = create_user_from_evecharacter_with_access(
            1001,
        )
        cls.audit = add_characteraudit_character_to_user(cls.user, 1001)

        cls.character_ownership.delete()

    def test_disable_characters_with_no_ownership_should_disable(self):
        # given
        scetions = CharacterAudit.UpdateSection.get_sections()
        for section in scetions:
            create_update_status(
                self.audit,
                section=section,
                is_success=True,
                error_message="",
                has_token_error=False,
                last_run_at=timezone.now(),
                last_run_finished_at=timezone.now(),
                last_update_at=timezone.now(),
                last_update_finished_at=timezone.now(),
            )

        # then
        self.assertEqual(CharacterAudit.objects.disable_characters_with_no_owner(), 1)


class TestCharacterAnnotateTotalUpdateStatus(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()
        load_eveentity()
        cls.user, cls.character_ownership = create_user_from_evecharacter_with_access(
            1001,
        )

    def test_should_be_ok(self):
        # given
        character = add_characteraudit_character_to_user(self.user, 1001)
        sections = CharacterAudit.UpdateSection.get_sections()
        for section in sections:
            create_update_status(
                character,
                section=section,
                is_success=True,
                error_message="",
                has_token_error=False,
                last_run_at=timezone.now(),
                last_run_finished_at=timezone.now(),
                last_update_at=timezone.now(),
                last_update_finished_at=timezone.now(),
            )

        # when
        update_status = CharacterAudit.objects.annotate_total_update_status()
        obj = update_status.first()

        # then
        self.assertEqual(obj.total_update_status, CharacterAudit.UpdateStatus.OK)

    def test_should_be_incomplete(self):
        # given
        character = add_characteraudit_character_to_user(self.user, 1001)
        sections = CharacterAudit.UpdateSection.get_sections()
        for section in sections[:2]:
            create_update_status(
                character,
                section=section,
                is_success=True,
                error_message="",
                has_token_error=False,
                last_run_at=timezone.now(),
                last_run_finished_at=timezone.now(),
                last_update_at=timezone.now(),
                last_update_finished_at=timezone.now(),
            )

        # when
        update_status = CharacterAudit.objects.annotate_total_update_status()
        obj = update_status.first()

        # then
        self.assertEqual(
            obj.total_update_status, CharacterAudit.UpdateStatus.INCOMPLETE
        )

    def test_should_be_token_error(self):
        # given
        character = add_characteraudit_character_to_user(self.user, 1001)
        sections = CharacterAudit.UpdateSection.get_sections()
        for section in sections:
            create_update_status(
                character,
                section=section,
                is_success=True,
                error_message="",
                has_token_error=False,
                last_run_at=timezone.now(),
                last_run_finished_at=timezone.now(),
                last_update_at=timezone.now(),
                last_update_finished_at=timezone.now(),
            )

        create_update_status(
            character,
            section=CharacterAudit.UpdateSection.WALLET_JOURNAL,
            is_success=False,
            error_message="",
            has_token_error=True,
            last_run_at=timezone.now(),
            last_run_finished_at=timezone.now(),
            last_update_at=timezone.now(),
            last_update_finished_at=timezone.now(),
        )

        # when
        update_status = CharacterAudit.objects.annotate_total_update_status()
        obj = update_status.first()

        # then
        self.assertEqual(
            obj.total_update_status, CharacterAudit.UpdateStatus.TOKEN_ERROR
        )

    def test_should_be_disabled(self):
        # given
        character = add_characteraudit_character_to_user(self.user, 1001)
        sections = CharacterAudit.UpdateSection.get_sections()
        for section in sections:
            create_update_status(
                character,
                section=section,
                is_success=True,
                error_message="",
                has_token_error=False,
                last_run_at=timezone.now(),
                last_run_finished_at=timezone.now(),
                last_update_at=timezone.now(),
                last_update_finished_at=timezone.now(),
            )

        character.active = False
        character.save()

        # when
        update_status = CharacterAudit.objects.annotate_total_update_status()
        obj = update_status.first()

        # then
        self.assertEqual(obj.total_update_status, CharacterAudit.UpdateStatus.DISABLED)

    def test_should_be_error(self):
        # given
        character = add_characteraudit_character_to_user(self.user, 1001)
        sections = CharacterAudit.UpdateSection.get_sections()
        for section in sections:
            create_update_status(
                character,
                section=section,
                is_success=False,
                error_message="",
                has_token_error=False,
                last_run_at=timezone.now(),
                last_run_finished_at=timezone.now(),
                last_update_at=timezone.now(),
                last_update_finished_at=timezone.now(),
            )

        # when
        update_status = CharacterAudit.objects.annotate_total_update_status()
        obj = update_status.first()

        # then
        self.assertEqual(obj.total_update_status, CharacterAudit.UpdateStatus.ERROR)



"""Django checks."""

# Standard Library
import numbers

# Django
from django.conf import settings
from django.core.checks import Warning as DjangoWarning
from django.core.checks import register


@register()
def ledger_config_check(app_configs, **kwargs):  # pylint: disable=unused-argument
    """ledger_config_check is a Django check that verifies that all periodic tasks
    have been configured correctly in settings.
    """
    if not hasattr(settings, "CELERYBEAT_SCHEDULE"):
        return []

    warnings = []
    for name, obj in settings.CELERYBEAT_SCHEDULE.items():
        _verify_task_config(
            warnings, name, obj, "ledger.tasks.update_subset_characters"
        )
        _verify_task_config(
            warnings, name, obj, "ledger.tasks.update_subset_corporations"
        )
    return warnings


def _verify_task_config(warnings: list, name: str, obj: dict, task_name: str):
    if obj["task"] != task_name:
        return

    schedule = obj["schedule"]
    if isinstance(schedule, numbers.Number):
        return

    warnings.append(
        DjangoWarning(
            (
                "Periodic task has deprecated schedule: "
                f'CELERYBEAT_SCHEDULE["{name}"]'
            ),
            hint=(
                'The value for "schedule" must be a positive number, '
                "not e.g. a cron definition. "
                f"Current value is: {schedule}"
            ),
            id="ledger.W001",
        )
    )



"""Initialize the app"""

__version__ = "0.9.9.3"
__title__ = "Ledger"

__package_name__ = "aa-ledger"

__app_name_useragent__ = "AA-Ledger"
__github_url__ = f"https://github.com/Geuthur/{__package_name__}"



# Generated by Django 4.2.23 on 2025-08-28 05:19

# Django
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("ledger", "0015_rename_character_characteraudit_eve_character"),
    ]

    operations = [
        migrations.AddField(
            model_name="characterminingledger",
            name="price_per_unit",
            field=models.DecimalField(
                decimal_places=2, default=None, max_digits=20, null=True
            ),
        ),
    ]



"""
Etag Helpers
"""

# Third Party
from bravado.exception import HTTPGatewayTimeout, HTTPNotModified

# Django
from django.core.cache import cache

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__
from ledger.app_settings import LEDGER_CACHE_KEY
from ledger.decorators import log_timing
from ledger.errors import HTTPGatewayTimeoutError, NotModifiedError

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

MAX_ETAG_LIFE = 60 * 60 * 24 * 7  # 7 Days


def get_etag_key(operation):
    """Get ETag Key"""
    return f"{LEDGER_CACHE_KEY}-{operation._cache_key()}"


def get_etag_header(operation):
    """Get Cached ETag"""
    return cache.get(get_etag_key(operation), False)


def del_etag_header(operation):
    """Delete Cached ETag"""
    return cache.delete(get_etag_key(operation), False)


def inject_etag_header(operation):
    """Inject ETag header"""
    etag = get_etag_header(operation)
    logger.debug(
        "ETag: get_etag %s - %s - etag:%s",
        operation.operation.operation_id,
        stringify_params(operation),
        etag,
    )
    if etag:
        operation.future.request.headers["If-None-Match"] = etag


def rem_etag_header(operation):
    """Remove ETag header"""
    logger.debug(
        "ETag: rem_etag %s - %s",
        operation.operation.operation_id,
        stringify_params(operation),
    )
    if "If-None-Match" in operation.future.request.headers:
        del operation.future.request.headers["If-None-Match"]
        return True
    return False


def set_etag_header(operation, headers):
    """Set ETag header"""
    etag_key = get_etag_key(operation)
    etag = headers.headers.get("ETag", None)
    if etag is not None:
        result = cache.set(etag_key, etag, MAX_ETAG_LIFE)
        logger.debug(
            "ETag: set_etag %s - %s - etag: %s - stored: %s",
            operation.operation.operation_id,
            stringify_params(operation),
            etag,
            result,
        )
        return True
    return False


def stringify_params(operation):
    """Stringify Operation Params"""
    out = []
    for p, v in operation.future.request.params.items():
        out.append(f"{p}: {v}")
    return ", ".join(out)


def handle_etag_headers(operation, headers, force_refresh, etags_incomplete):
    if (
        get_etag_header(operation) == headers.headers.get("ETag")
        and not force_refresh
        and not etags_incomplete
    ):
        # Etag Match Cache Check
        raise NotModifiedError()

    if force_refresh:
        # Remove Etag if force_refresh
        logger.debug(
            "ETag: Removing Etag %s F:%s - %s",
            operation.operation.operation_id,
            force_refresh,
            stringify_params(operation),
        )
        del_etag_header(operation)
    else:
        # Save Etag
        logger.debug(
            "ETag: Saving Etag %s F:%s - %s",
            operation.operation.operation_id,
            force_refresh,
            stringify_params(operation),
        )
        set_etag_header(operation, headers)


def handle_page_results(
    operation, current_page, total_pages, etags_incomplete, force_refresh
):
    results = []
    while current_page <= total_pages:
        operation.future.request.params["page"] = current_page
        try:
            if not etags_incomplete and not force_refresh:
                inject_etag_header(operation)
            else:
                rem_etag_header(operation)

            result, headers = operation.result()
            total_pages = int(headers.headers["X-Pages"])
            logger.debug(
                "Cached ETag: %s - ESI ETag: %s - Last-Modified: %s - Operation ID: %s",
                get_etag_header(operation),
                headers.headers.get("ETag"),
                headers.headers.get("Last-Modified"),
                operation.operation.operation_id,
            )

            handle_etag_headers(operation, headers, force_refresh, etags_incomplete)

            # Store results
            results += result
            current_page += 1

            if not etags_incomplete and not force_refresh:
                logger.debug(
                    "ETag: No Etag: %s - %s",
                    operation.operation.operation_id,
                    stringify_params(operation),
                )
                current_page = 1
                results = []
                etags_incomplete = True

        except (HTTPNotModified, NotModifiedError) as e:
            try:
                etag = e.response.headers["ETag"]
            except AttributeError:
                etag = None

            if isinstance(e, NotModifiedError):
                logger.debug(
                    "ETag: Match Cache - Etag: %s, %s",
                    etag,
                    stringify_params(operation),
                )
                total_pages = int(headers.headers["X-Pages"])
            else:
                logger.debug(
                    "ETag: Match ESI - Etag: %s - %s ETag-Incomplete: %s",
                    etag,
                    stringify_params(operation),
                    etags_incomplete,
                )
                total_pages = int(e.response.headers["X-Pages"])

            if not etags_incomplete:
                current_page += 1
            else:
                current_page = 1
                results = []
                etags_incomplete = False

    if not etags_incomplete and not force_refresh:
        raise NotModifiedError()

    return results, current_page, total_pages


@log_timing(logger)
def etag_results(operation, token, force_refresh=False):
    """Handle ETag results"""
    logger.debug(
        "ETag: etag_results %s - %s",
        operation.operation.operation_id,
        force_refresh,
    )
    operation.request_config.also_return_response = True
    if token:
        operation.future.request.headers["Authorization"] = (
            "Bearer " + token.valid_access_token()
        )
    if "page" in operation.operation.params:
        current_page = 1
        total_pages = 1
        etags_incomplete = False

        try:
            results, current_page, total_pages = handle_page_results(
                operation, current_page, total_pages, etags_incomplete, force_refresh
            )
        except HTTPGatewayTimeout as e:
            logger.debug("ETag: Gateway Timeout %s", operation.operation.operation_id)
            raise HTTPGatewayTimeoutError() from e
        except NotModifiedError as e:
            logger.debug("ETag: Not Modified %s", operation.operation.operation_id)
            raise NotModifiedError() from e
    else:
        if not force_refresh:
            inject_etag_header(operation)
        try:
            results, headers = operation.result()
        except HTTPNotModified as e:
            logger.debug("ETag: HTTP Not Modified %s", operation.operation.operation_id)
            set_etag_header(operation, e.response)
            raise NotModifiedError() from e
        except HTTPGatewayTimeout as e:
            logger.debug("ETag: Gateway Timeout %s", operation.operation.operation_id)
            raise HTTPGatewayTimeoutError() from e
        handle_etag_headers(operation, headers, force_refresh, etags_incomplete=False)
    return results



"""
Core View Helper
"""

# Standard Library
from decimal import Decimal
from hashlib import md5

# Django
from django.db.models import Q, QuerySet, Sum
from django.urls import reverse
from django.utils import timezone
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _

# Alliance Auth
from allianceauth.authentication.models import CharacterOwnership, UserProfile
from allianceauth.eveonline.models import (
    EveAllianceInfo,
    EveCharacter,
    EveCorporationInfo,
)
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__
from ledger.api.api_helper.billboard_helper import BillboardSystem
from ledger.app_settings import LEDGER_CACHE_KEY
from ledger.helpers.ref_type import RefTypeManager
from ledger.models.general import EveEntity

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def add_info_to_context(request, context: dict) -> dict:
    """Add additional information to the context for the view."""
    # pylint: disable=import-outside-toplevel
    # AA Ledger
    from ledger.models.characteraudit import CharacterAudit

    total_issues = (
        CharacterAudit.objects.annotate_total_update_status_user(user=request.user)
        .aggregate(total_failed=Sum("num_sections_failed"))
        .get("total_failed", 0)
    )

    new_context = {
        **{
            "issues": total_issues,
        },
        **context,
    }
    return new_context


class DummyEveEntity:
    """Dummy Eve Entity class for fallback when no entity is found."""

    def __init__(self, entity_id, entity_name="Unknown"):
        self.entity_id = entity_id
        self.entity_name = entity_name
        self.type = "character"


class LedgerEntity:
    """Class to hold character or corporation data for the ledger."""

    # pylint: disable=too-many-positional-arguments
    def __init__(
        self,
        entity_id,
        character_obj: EveCharacter = None,
        corporation_obj: EveCorporationInfo = None,
        alliance_obj: EveAllianceInfo = None,
        details_url=None,
    ):
        self.type = "character"
        self.entity = None
        self.entity_id = entity_id
        self.entity_name = None
        self.details_url = details_url
        if character_obj and hasattr(character_obj, "character_id"):
            self.entity = character_obj
            self.entity_id = character_obj.character_id
            self.entity_name = character_obj.character_name
        elif corporation_obj and hasattr(corporation_obj, "corporation_id"):
            self.entity = corporation_obj
            self.entity_id = corporation_obj.corporation_id
            self.entity_name = corporation_obj.corporation_name
            self.type = "corporation"
        elif alliance_obj and hasattr(alliance_obj, "alliance_id"):
            self.entity = alliance_obj
            self.entity_id = alliance_obj.alliance_id
            self.entity_name = alliance_obj.alliance_name
            self.type = "alliance"
        else:
            try:
                entity_obj = EveEntity.objects.get(eve_id=entity_id)
                self.entity = entity_obj
                self.entity_id = entity_obj.eve_id
                self.entity_name = entity_obj.name
                self.type = entity_obj.category
            except EveEntity.DoesNotExist:
                self.entity = DummyEveEntity(entity_id, "Unknown")
                self.entity_id = entity_id
                self.entity_name = "Unknown"

    @property
    def is_eve_character(self):
        """Check if the entity is an Eve Character."""
        return isinstance(self.entity, EveCharacter)

    @property
    def is_eve_corporation(self):
        """Check if the entity is an Eve Corporation."""
        return isinstance(self.entity, EveCorporationInfo)

    @property
    def is_eve_entity(self):
        """Check if the entity is an Eve Entity."""
        return isinstance(self.entity, EveEntity)

    @property
    def alts(self) -> QuerySet[EveCharacter]:
        """Get all alts for this character."""
        if not isinstance(self.entity, EveCharacter):
            raise ValueError("Entity is not an EveCharacter.")
        alts = EveCharacter.objects.filter(
            character_ownership__user=self.entity.character_ownership.user
        ).select_related(
            "character_ownership",
        )
        return alts

    def portrait_url(self):
        """Return the portrait URL for the entity."""
        try:
            if isinstance(self.entity, EveCorporationInfo):
                return self.entity.logo_url_32

            if hasattr(self.entity, "portrait_url"):
                return self.entity.portrait_url(size=32)

            if self.entity.category == "faction":
                return ""
            return self.entity.icon_url(size=32)
        except Exception as e:  # pylint: disable=broad-except
            logger.error(f"Error getting portrait URL for {self.entity_id}: {e}")
            return ""

    def add_details_url(self, details_url):
        """Set the details URL for the entity."""
        self.details_url = details_url

    def get_alts_ids_or_self(self):
        """Return the IDs of all alternative characters or the character ID itself."""
        try:
            character = EveCharacter.objects.get(character_id=self.entity_id)
            if hasattr(character, "character_ownership"):
                alt_ids = character.character_ownership.user.character_ownerships.all().values_list(
                    "character__character_id", flat=True
                )
                return list(alt_ids)
        except (
            EveCharacter.DoesNotExist,
            AttributeError,
            CharacterOwnership.DoesNotExist,
        ):
            pass
        return [self.entity_id]

    @property
    def create_button(self):
        """Generate the URL for character details."""
        title = _("View Details")
        button_html = f"<button class='btn btn-primary btn-sm btn-square' data-bs-toggle='modal' data-bs-target='#modalViewCharacterContainer' data-ajax_url='{self.details_url}' title='{title}' data-tooltip-toggle='ledger-tooltip'><span class='fas fa-info'></span></button>"
        return mark_safe(button_html)


class LedgerCore:
    """Core View Helper for Ledger."""

    def __init__(self, year=None, month=None, day=None):
        self.date_info = {"year": year, "month": month, "day": day}
        self.ledger_type = "ledger"

        # If all are None, default to 'month' view
        if year is None and month is None and day is None:
            self.view = "month"
        else:
            self.view = "day" if day else "month" if month else "year"

        self.journal = None
        self.mining = None
        self.billboard = BillboardSystem(self.view)

    @property
    def year(self):
        return self.date_info["year"]

    @property
    def month(self):
        return self.date_info["month"]

    @property
    def day(self):
        return self.date_info["day"]

    @property
    def filter_date(self):
        """
        Generate a date filter for the ledger based on year, month, and day.
        Returns:
            Q: A Django Q object representing the date filter.
        """
        now = timezone.now()
        # If all are None, use current year and month
        if self.year is None and self.month is None and self.day is None:
            filter_date = Q(date__year=now.year) & Q(date__month=now.month)
        else:
            filter_date = (
                Q(date__year=self.year) if self.year else Q(date__year=now.year)
            )
            if self.month:
                filter_date &= Q(date__month=self.month)
            if self.day:
                filter_date &= Q(date__day=self.day)
        return filter_date

    @property
    def get_details_title(self):
        """
        Generate a title for the details view based on the date information.

        Returns:
            str: A formatted string representing the date or a default title.
        """
        if self.year and self.month and self.day:
            return f"{self.year:04d}-{self.month:02d}-{self.day:02d}"
        if self.year and self.month:
            return f"{self.year:04d}-{self.month:02d}"
        if self.year:
            return f"{self.year:04d}"
        return "Character Ledger Details"

    @property
    def auth_accounts(self):
        """Get all user accounts with a main character."""
        return (
            UserProfile.objects.filter(
                main_character__isnull=False,
            )
            .prefetch_related(
                "user__profile__main_character",
            )
            .order_by(
                "user__profile__main_character__character_name",
            )
        )

    @property
    def auth_character_ids(self) -> set:
        """Get all account Character IDs from Alliance Auth."""
        account_character_ids = set()
        for account in self.auth_accounts:
            alts = account.user.character_ownerships.all()
            account_character_ids.update(
                alts.values_list("character__character_id", flat=True)
            )
        return account_character_ids

    def _get_ledger_journal_hash(self, journal: list[str]) -> str:
        """Generate a hash for the ledger journal."""
        return md5(",".join(str(x) for x in sorted(journal)).encode()).hexdigest()

    def _get_ledger_hash(self, header_key: str) -> str:
        """Generate a hash for the ledger journal."""
        return md5(header_key.encode()).hexdigest()

    def _get_ledger_header(
        self, ledger_args: str, year: int, month: int, day: int
    ) -> str:
        """Generate a header string for the ledger."""
        return f"{ledger_args}_{year}_{month}_{day}"

    def _build_ledger_cache_key(self, header_key: str) -> str:
        """Build a cache key for the ledger."""
        return f"{LEDGER_CACHE_KEY}-{self._get_ledger_hash(header_key)}"

    def _calculate_totals(self, ledger) -> dict:
        """
        Calculate the total amounts for each category in the ledger.

        Args:
            ledger (list or dict): The ledger data to calculate totals from.

        Returns:
            dict: A dictionary containing the totals for each category.
        """
        totals = {
            "bounty": Decimal(0),
            "ess": Decimal(0),
            "costs": Decimal(0),
            "mining": Decimal(0),
            "miscellaneous": Decimal(0),
            "total": Decimal(0),
        }

        if not ledger:
            return totals

        if isinstance(ledger, dict):
            ledger = [ledger]

        for total in ledger:

            if total is None:
                continue

            totals["bounty"] += total["ledger"].get("bounty", 0)
            totals["ess"] += total["ledger"].get("ess", 0)
            totals["costs"] += total["ledger"].get("costs", 0)
            totals["mining"] += total["ledger"].get("mining", 0)
            totals["miscellaneous"] += total["ledger"].get("miscellaneous", 0)
            totals["total"] += total["ledger"].get("total", 0)
        return totals

    def create_url(self, viewname: str, **kwargs):
        """
        Create a URL for the given view and entity using kwargs.
        Args:
            viewname: The name of the view to create the URL for.
            kwargs: All needed parameters for the URL (e.g. character_id, corporation_id, etc.)
        Returns:
            A URL string for the specified view.
        """
        if self.year and self.month and self.day:
            return reverse(
                f"ledger:{viewname}",
                kwargs={
                    **kwargs,
                    "year": self.year,
                    "month": self.month,
                    "day": self.day,
                },
            )
        if self.year and self.month:
            return reverse(
                f"ledger:{viewname}",
                kwargs={
                    **kwargs,
                    "year": self.year,
                    "month": self.month,
                },
            )
        if self.year:
            return reverse(
                f"ledger:{viewname}",
                kwargs={**kwargs, "year": self.year},
            )
        return reverse(
            f"ledger:{viewname}",
            kwargs={
                **kwargs,
                "year": timezone.now().year,
                "month": timezone.now().month,
            },
        )

    def create_view_data(self, viewname: str, **kwargs) -> dict:
        """
        Create view data for the ledger using kwargs.
        Args:
            viewname (str): The name of the view to create the URL for.
            kwargs: All needed parameters for the URL (e.g. character_id, corporation_id, etc.)
        Returns:
            dict: A dictionary containing the type, date, and details URL.
        """
        return {
            "type": self.ledger_type,
            "date": {
                "current": {
                    "year": timezone.now().year,
                    "month": timezone.now().month,
                    "day": timezone.now().day,
                },
                "year": self.year,
                "month": self.month,
                "day": self.day,
            },
            "details_url": self.create_url(
                viewname=viewname,
                **kwargs,
            ),
        }

    # pylint: disable=too-many-locals
    def _generate_amounts(
        self,
        income_types: list,
        entity: LedgerEntity,
        is_old_ess: bool = False,
        char_ids: list = None,
    ) -> dict:
        """Generate amounts for the entity based on income types and reference types."""
        amounts = {}

        ref_types = RefTypeManager.get_all_categories()

        # Bounty Income
        if not entity.entity_id == 1000125:  # Remove Concord Bountys
            bounty_income = self.journal.aggregate_bounty()
            if bounty_income > 0:
                amounts["bounty_income"] = {
                    "total_amount": bounty_income,
                    "ref_types": ["bounty_prizes"],
                }

        if isinstance(entity.entity, EveCharacter):
            # Mining Income
            mining_income = self.mining.aggregate_mining()
            if mining_income > 0:
                amounts["mining_income"] = {
                    "total_amount": mining_income,
                    "ref_types": ["mining"],
                }

        # ESS Income (nur wenn bounty_income existiert)
        ess_income = (
            bounty_income * Decimal(0.667)
            if is_old_ess and bounty_income
            else self.journal.aggregate_ess()
        )
        if ess_income > 0:
            amounts["ess_income"] = {
                "total_amount": ess_income,
                "ref_types": ["ess_escrow_transfer"],
            }

        # Income/Cost Ref Types (DRY)
        for ref_type, value in ref_types.items():
            ref_type_name = ref_type.lower()
            for kind, income_flag in (("income", True), ("cost", False)):
                kwargs = {"ref_type": value, "income": income_flag}
                kwargs = RefTypeManager.special_cases_details(
                    value, entity, kwargs, journal_type=entity.type, char_ids=char_ids
                )
                agg = self.journal.aggregate_ref_type(**kwargs)
                if (income_flag and agg > 0) or (not income_flag and agg < 0):
                    amounts[f"{ref_type_name}_{kind}"] = {
                        "total_amount": agg,
                        "ref_types": value,
                    }

        # Summary (exclude mining_income)
        summary = sum(
            amount["total_amount"]
            for key, amount in amounts.items()
            if key != "mining_income"
            and isinstance(amount, dict)
            and "total_amount" in amount
        )

        if summary == 0:
            return None

        amounts["summary"] = {
            "total_amount": summary,
        }

        # Dynamische Income/Cost-Typen für das Template
        income_types += [
            (f"{ref_type.lower()}_income", _(ref_type.replace("_", " ").title()))
            for ref_type in ref_types
        ]
        cost_types = [
            (f"{ref_type.lower()}_cost", _(ref_type.replace("_", " ").title()))
            for ref_type in ref_types
        ]
        amounts["income_types"] = income_types
        amounts["cost_types"] = cost_types
        return amounts

    def _create_corporation_details(self, entity: LedgerEntity) -> dict:
        """Create the corporation amounts for the Information View."""
        # NOTE (can only used if setup_ledger is defined in the subclass)
        self.setup_ledger(entity=entity)  # pylint: disable=no-member

        income_types = [
            ("bounty_income", _("Bounty")),
            ("ess_income", _("Encounter Surveillance System")),
        ]
        amounts = self._generate_amounts(income_types=income_types, entity=entity)
        return amounts

    # pylint: disable=no-member
    def _create_character_details(self) -> dict:
        """
        Create the character amounts for the Information View.
        Only work with CharacterData Class
        """
        if not self.character:
            raise ValueError("No Character Data found.")

        # NOTE (can only used if setup_ledger is defined in the subclass)
        self.setup_ledger(self.character)

        entity = LedgerEntity(
            entity_id=self.character.eve_character.character_id,
            character_obj=self.character.eve_character,
        )

        income_types = [
            ("bounty_income", _("Bounty")),
            ("ess_income", _("Encounter Surveillance System")),
            ("mining_income", _("Mining")),
        ]
        amounts = self._generate_amounts(
            income_types=income_types,
            entity=entity,
            is_old_ess=self.is_old_ess,
            char_ids=self.alts_ids,
        )
        return amounts

    def _add_average_details(self, request, amounts, day: int = None):
        """Add average details to the amounts dictionary, skipping if no data or total is 0."""
        if amounts is None:
            return None

        avg = day if day else timezone.now().day
        if request.GET.get("all", False):
            avg = 365

        for key in amounts:
            if (
                isinstance(amounts[key], dict)
                and "total_amount" in amounts[key]
                and amounts[key]["total_amount"] not in (None, 0, 0.0, Decimal(0))
            ):
                total = amounts[key]["total_amount"]
                amounts[key]["average_day"] = total / avg
                amounts[key]["average_hour"] = total / avg / 24
                amounts[key]["average_tick"] = total / 20
                amounts[key]["current_day_tick"] = (
                    amounts[key].get("total_amount_day", 0) / 20
                )
                amounts[key]["average_day_tick"] = total / avg / 20
                amounts[key]["average_hour_tick"] = total / avg / 24 / 20
        return amounts

    def _build_xy_chart(self, title: str):
        """Build the XY chart for the billboard."""
        if not self.billboard.results:
            return

        xy_data, categories = self.billboard.generate_xy_series()
        self.billboard.create_xy_chart(
            title=title,
            categories=categories,
            series=xy_data,
        )
